use std::collections::HashMap;

use gloo_console::log;
use gloo_net::http::Response;
use itertools::Itertools;
use yew::{prelude::*, Component, Properties};

use crate::{
    filter_menu::Filter,
    requests,
    types::{
        Benutzer, GeplanteVpZuordnung, JahresVpZuordnung, Verteilpunkt, VpWunsch, VpZuordnung,
        Wünsche,
    },
};
/// An enumeration representing different types of data that can be received from the server.
pub enum NetworkObjekte {
    /// Distribution point data.
    Verteilpunkt(HashMap<i32, Verteilpunkt>),
    /// Wish data.
    Wunsch(Option<Wünsche>),
    /// User data.
    User(Option<HashMap<i32, Benutzer>>),
    /// Previous year's assignment data.
    ZuordnungVorjahr(Option<JahresVpZuordnung>),
}

/// An enumeration representing messages that can be passed between the model and the view.
pub enum ZuordnungMessage {
    /// A message requesting data from the server.
    Get,
    /// A message indicating that data has been received from the server.
    Recieved(NetworkObjekte, Result<Response, gloo_net::Error>),
    /// A message indicating that data has been parsed.
    Parsed(NetworkObjekte),
    /// A message indicating that the year has been updated.
    UpdatedJahr(Option<u32>),
    /// A message indicating that a dialog should be displayed.
    ShowDialog(Dialog),
    /// A message indicating that the current dialog should be hidden.
    HideDialog,
}

pub struct Dialog {
    pub title: String,
    pub content: DialogContent,
}

pub enum DialogContent {
    Wunschliste(Vec<VpWunsch>),
}

#[derive(Properties, PartialEq, Default)]
pub struct ZuordnungProps {}

/// A struct representing the state of the assignment app.
pub struct ZuordnungApp {
    /// The year for which assignments are being made.
    jahr: u32,
    /// A mapping from distribution point IDs to distribution point objects.
    verteilpunkte: HashMap<i32, Verteilpunkt>,
    /// The list of wishes made by users.
    wünsche: Option<Wünsche>,
    /// A mapping from user IDs to user objects.
    user: Option<HashMap<i32, Benutzer>>,
    /// The assignments made for the previous year.
    zuordnung_vorjahr: Option<JahresVpZuordnung>,
    /// The planned assignments for the current year.
    geplante_zuordnung: HashMap<i32, GeplanteVpZuordnung>,
    /// The currently displayed dialog.
    dialog: Option<Dialog>,
}

impl Component for ZuordnungApp {
    type Message = ZuordnungMessage;

    type Properties = ZuordnungProps;

    fn create(ctx: &yew::Context<Self>) -> Self {
        ctx.link().send_message(ZuordnungMessage::Get);
        Self {
            jahr: 0,
            verteilpunkte: HashMap::new(),
            wünsche: None,
            user: None,
            zuordnung_vorjahr: None,
            geplante_zuordnung: HashMap::new(),
            dialog: None,
        }
    }

    fn update(&mut self, ctx: &Context<Self>, msg: Self::Message) -> bool {
        match msg {
            ZuordnungMessage::Get => {
                ctx.link().send_future(requests::fetch_verteilpunkte());
                log!("Update der Verteilpunkte");
                false
            }
            ZuordnungMessage::Recieved(t, e) => {
                ctx.link().send_future(async move {
                    let network_objekt = match t {
                        NetworkObjekte::Verteilpunkt(_) => {
                            let response = e.unwrap().json().await;
                            NetworkObjekte::Verteilpunkt(response.unwrap())
                        }
                        NetworkObjekte::Wunsch(_) => {
                            let response = e.unwrap().json().await;
                            NetworkObjekte::Wunsch(Some(response.unwrap()))
                        }
                        NetworkObjekte::User(_) => {
                            let response = e.unwrap().json().await;
                            NetworkObjekte::User(Some(response.unwrap()))
                        }
                        NetworkObjekte::ZuordnungVorjahr(_) => {
                            let response = e.unwrap().json().await;
                            NetworkObjekte::ZuordnungVorjahr(Some(response.unwrap()))
                        }
                    };
                    ZuordnungMessage::Parsed(network_objekt)
                });
                log!("received");
                true
            }
            ZuordnungMessage::UpdatedJahr(jahr) => {
                log!("Test updated to {}", jahr);
                if let Some(jahr) = jahr {
                    self.jahr = jahr;
                    ctx.link().send_future(requests::fetch_wünsche(jahr));
                    ctx.link().send_future(requests::fetch_user(jahr));
                    ctx.link().send_future(requests::fetch_zuordnung(jahr - 1));
                    true
                } else {
                    false
                }
            }
            ZuordnungMessage::Parsed(t) => match t {
                NetworkObjekte::Verteilpunkt(v) => {
                    log!("Verteilpunkte geparsed");
                    self.verteilpunkte = v;
                    true
                }
                NetworkObjekte::Wunsch(v) => {
                    log!("Wünsche geparsed!");
                    self.wünsche = v;
                    true
                }
                NetworkObjekte::User(v) => {
                    log!("User geparsed");
                    self.user = v;
                    true
                }
                NetworkObjekte::ZuordnungVorjahr(v) => {
                    log!("Vorjahreszuordnung geparsed");
                    self.zuordnung_vorjahr = v;
                    true
                }
            },
            ZuordnungMessage::ShowDialog(dialog) => {
                self.dialog = Some(dialog);
                true
            }
            ZuordnungMessage::HideDialog => {
                self.dialog = None;
                true
            }
        }
    }

    fn view(&self, ctx: &yew::Context<Self>) -> yew::Html {
        let callback_jahr = ctx.link().callback(ZuordnungMessage::UpdatedJahr);

        html! {<>
        <h1>{"Verteilpunktzuordnung"}{if self.jahr>2000 {html!{<>{" für das Jahr "} {self.jahr}</>}}else{html!{}}}</h1>
        <div class="filter"><Filter {callback_jahr}/></div>
        {
            if self.wünsche.is_some(){html!{
            <div class="table-responsive">
                <table class="table table-striped table-hover">
                    <thead class="table-dark">
                        {self.tabellen_überschrift(&self.wünsche)}
                    </thead>
                    <tbody>
                {self.verteilpunkte
                    .iter()
                    .map(|(id, v)| html! {
                        <tr>
                            <td>{self.verteilpunkt_name(v)}</td>
                            <td>{self.aktuelle_mitglieder(v, &self.zuordnung_vorjahr)}</td>
                            <td>{self.zukünftige_mitglieder(v)}</td>
                            {self.count_wünsche(*id, ctx)}
                        </tr>
                    })
                    .collect::<Html>()}
                </tbody></table>
                {{self.show_dialog( ctx)}}
            </div>}}else{html!{<h3 style="padding-top: 100px; text-align:center;">{"Es ist keine Saison ausgewählt"}</h3>}}}</>
        }
    }
}
impl ZuordnungApp {
    fn count_wünsche(&self, verteilpunkt_id: i32, ctx: &yew::Context<ZuordnungApp>) -> Html {
        // Erst wenn die Wünsche geladen sind wird etwas angezeigt.
        if let Some(wünsche) = &self.wünsche {
            // zähle nur die wünsche für den aktuellen Verteilpunkt
            let filtered = wünsche
                .wünsche
                .iter()
                .filter(|(_, wunsch)| wunsch.verteilpunkt_id == verteilpunkt_id);
            // A mapping from priority values to lists of wishes.
            let counts: HashMap<i32, Vec<VpWunsch>> = wünsche
                .prioritäten
                .iter()
                .map(|prio| (*prio, Vec::new()))
                .collect();
            let counts: HashMap<i32, Vec<VpWunsch>> =
                filtered.fold(counts, |mut counts, (_, wunsch)| {
                    counts
                        .entry(wunsch.prioritaet)
                        .or_insert_with(Vec::new)
                        .push(wunsch.clone());
                    counts
                });
            // sortiere nach Priorität
            let sorted_and_grouped_by_priority = counts.into_iter().sorted_by_key(|x| x.0);

            sorted_and_grouped_by_priority
                .into_iter()
                .map(|(_priorität, wunsch)| {
                    let length = wunsch.len();
                    let verteilpunkte = self.verteilpunkte.clone();

                    let onclick = ctx.link().callback(move |_| {
                        ZuordnungMessage::ShowDialog(Dialog {
                            title: {
                                let vp = verteilpunkte
                                    .get(&verteilpunkt_id)
                                    .expect("should allways be there");
                                format!("Wünsche für {}", vp.name)
                            },
                            content: DialogContent::Wunschliste(wunsch.clone()),
                        })
                    });
                    html! {<td><b {onclick}>{length}</b></td>}
                })
                .collect()
        } else {
            html!()
        }
    }

    fn tabellen_überschrift(&self, wünsche: &Option<Wünsche>) -> Html {
        // Generate the HTML for the wish priorities
        fn wünsche_überschrift(prioritäten: &[i32]) -> Html {
            prioritäten
                .iter()
                .map(|id| html! {<th>{id}{". Wunsch"}</th>})
                .collect()
        }

        match wünsche {
            Some(wünsche) => {
                let wünsche_überschrift = wünsche_überschrift(&wünsche.prioritäten);

                // Render the table header
                html! {
                    <>
                    <tr>
                      <th scope="col">{"Verteilpunkt"}</th>
                      <th scope="col" colspan="2">{"Mitglieder"}</th>
                      <th scope="col" colspan={wünsche.prioritäten.len().to_string()}>{"Nicht zugeordnete Bieter"}</th>
                    </tr>
                    <tr>
                      <th scope="col"></th>
                      <th scope="col">{"aktuell"}</th>
                      <th scope="col">{"zukünftig"}</th>
                      {wünsche_überschrift}
                    </tr></>
                }
            }
            None => html!(),
        }
    }

    fn verteilpunkt_name(&self, verteilpunkt: &Verteilpunkt) -> Html {
        // Render a div with a checkbox and label
        html! {
            <div class="form-check form-switch">
                <input class="form-check-input"
                       type="checkbox"
                       role="switch"
                       id="flexSwitchCheckChecked2"
                       checked=true />
                <label class="form-check-label" for="flexSwitchCheckChecked">{&verteilpunkt.name}</label>
            </div>
        }
    }

    fn aktuelle_mitglieder(&self, v: &Verteilpunkt, z: &Option<JahresVpZuordnung>) -> Html {
        // Count the number of items in the zuordnungen map with the matching verteilpunkt_id
        fn count_zuordnungen(v: &Verteilpunkt, zuordnungen: &HashMap<i32, VpZuordnung>) -> usize {
            zuordnungen
                .iter()
                .filter(|(_, zuo)| v.id == zuo.verteilpunkt_id)
                .count()
        }

        match z {
            Some(z) => {
                // Render the count and the kapazitaet as "X/Y"
                let count = count_zuordnungen(v, &z.zuordnungen);
                html! {<>{count}{"/"} {v.kapazitaet}</>}
            }
            None => html!(),
        }
    }

    fn zukünftige_mitglieder(&self, verteilpunkt: &Verteilpunkt) -> Html {
        // Count the number of items in the z map with the matching verteilpunkt_id
        let count = &self
            .geplante_zuordnung
            .iter()
            .filter(|(_, zuo)| verteilpunkt.id == zuo.verteilpunkt_id)
            .count();

        // Render the count and the kapazitaet as "X/Y"
        html! {<>{count}{"/"} {verteilpunkt.kapazitaet}</>}
    }

    fn show_dialog(&self, ctx: &yew::Context<Self>) -> Html {
        match &self.dialog {
            Some(dialog) => {
                // Create a callback that calls the HideDialog function when triggered
                let onclick = ctx.link().callback(|_| ZuordnungMessage::HideDialog);

                html! {
                    // Render the dialog container with the onclick callback attached
                    <div class="modal_wasm" {onclick}>
                      <div class="container">
                        <div class="body">
                            // Render the dialog title
                            <h3>{&dialog.title}</h3>
                            // Render the dialog content
                            {self.show_dialog_content(&dialog.content)}
                        </div>
                      </div>
                    </div>
                }
            }
            None => html! {},
        }
    }

    fn show_dialog_content(&self, content: &DialogContent) -> Html {
        match content {
            DialogContent::Wunschliste(content) => {
                let mut htmlnodes = Vec::new();

                // Iterate over each wish in the list
                for wunsch in content {
                    // Find the user associated with this wish
                    let user = match &self.user {
                        Some(users) => match users.get(&wunsch.user_id) {
                            Some(user) => user,
                            None => continue, // Skip this wish if the user is not found
                        },
                        None => continue, // Skip this wish if self.user is not set
                    };

                    // Append the user's first and last name to the HTML
                    htmlnodes
                        .push(html! {<tr><td>{&user.vorname}</td><td>{&user.nachname}</td></tr>});
                }

                html! {{ htmlnodes.into_iter().collect::<Html>() }}
            }
        }
    }
}
